# -*- coding: utf-8 -*-
import pathlib
import os

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import eyetracker

from .adhawk_mindlink import preprocessData as adhawk_mindlink
from .SeeTrue_STONE import preprocessData as SeeTrue_STONE
from .SMI_ETG import preprocessData as SMI_ETG
from .tobii_G2 import preprocessData as tobii_G2
from .tobii_G3 import preprocessData as tobii_G3

def pupil_core(output_dir: str | pathlib.Path, source_dir: str | pathlib.Path = None, rec_info: Recording = None, copy_scene_video = True) -> Recording:
    from .pupilLabs import preprocessData
    return preprocessData(output_dir, 'Pupil Core', source_dir, rec_info, copy_scene_video)

def pupil_invisible(output_dir: str | pathlib.Path, source_dir: str | pathlib.Path = None, rec_info: Recording = None, copy_scene_video = True) -> Recording:
    from .pupilLabs import preprocessData
    return preprocessData(output_dir, 'Pupil Invisible', source_dir, rec_info, copy_scene_video)

def pupil_neon(output_dir: str | pathlib.Path, source_dir: str | pathlib.Path = None, rec_info: Recording = None, copy_scene_video = True) -> Recording:
    from .pupilLabs import preprocessData
    return preprocessData(output_dir, 'Pupil Neon', source_dir, rec_info, copy_scene_video)


def get_recording_info(source_dir: str | pathlib.Path, device: str | EyeTracker) -> list[Recording]:
    source_dir  = pathlib.Path(source_dir)
    device = eyetracker.string_to_enum(device)

    rec_info = None
    match device:
        case EyeTracker.Pupil_Core:
            from .pupilLabs import getRecordingInfo
            rec_info = getRecordingInfo(source_dir, device)
        case EyeTracker.Pupil_Invisible:
            from .pupilLabs import getRecordingInfo
            rec_info = getRecordingInfo(source_dir, device)
        case EyeTracker.Pupil_Neon:
            from .pupilLabs import getRecordingInfo
            rec_info = getRecordingInfo(source_dir, device)
        case EyeTracker.SeeTrue_STONE:
            from .SeeTrue_STONE import getRecordingInfo
            rec_info = getRecordingInfo(source_dir)
        case EyeTracker.SMI_ETG:
            from .SMI_ETG import getRecordingInfo
            rec_info = getRecordingInfo(source_dir)
        case EyeTracker.Tobii_Glasses_2:
            from .tobii_G2 import getRecordingInfo
            rec_info = getRecordingInfo(source_dir)
        case EyeTracker.Tobii_Glasses_3:
            from .tobii_G3 import getRecordingInfo
            rec_info = getRecordingInfo(source_dir)
        case EyeTracker.AdHawk_MindLink:
            from .adhawk_mindlink import getRecordingInfo
            rec_info = getRecordingInfo(source_dir)

    if rec_info is not None and not isinstance(rec_info,list):
        rec_info = [rec_info]
    return rec_info


# single front end to the various device import functions for convenience
def do_import(output_dir: str | pathlib.Path = None, source_dir: str | pathlib.Path = None, device: str | EyeTracker = None, rec_info: Recording = None, copy_scene_video = True) -> Recording:
    # output_dir is the working directory folder where the export of this recording will be placed
    # should match rec_info.working_directory if both are provided (is checked below)
    if rec_info is not None:
        if isinstance(rec_info,list):
            raise ValueError('You should provide a single Recording to this function''s "rec_info" input argument, not a list of Recordings.')
    device    , rec_info = check_device    (device    , rec_info)
    source_dir, rec_info = check_source_dir(source_dir, rec_info)
    output_dir, rec_info = check_output_dir(output_dir, rec_info)

    # do the actual import/pre-process
    match device:
        case EyeTracker.AdHawk_MindLink:
            rec_info = adhawk_mindlink(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)
        case EyeTracker.Pupil_Core:
            rec_info = pupil_core(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)
        case EyeTracker.Pupil_Invisible:
            rec_info = pupil_invisible(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)
        case EyeTracker.Pupil_Neon:
            rec_info = pupil_neon(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)
        case EyeTracker.SeeTrue_STONE:
            rec_info = SeeTrue_STONE(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)
        case EyeTracker.SMI_ETG:
            rec_info = SMI_ETG(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)
        case EyeTracker.Tobii_Glasses_2:
            rec_info = tobii_G2(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)
        case EyeTracker.Tobii_Glasses_3:
            rec_info = tobii_G3(output_dir, source_dir, rec_info, copy_scene_video=copy_scene_video)

    return rec_info


def check_source_dir(source_dir: str|pathlib.Path, rec_info: Recording) -> tuple[pathlib.Path, Recording]:
    if source_dir is not None:
        source_dir  = pathlib.Path(source_dir)
        if rec_info is not None and rec_info.source_directory and pathlib.Path(rec_info.source_directory) != source_dir:
            raise ValueError(f"The provided source_dir ({source_dir}) does not equal the source directory set in rec_info ({rec_info.source_directory}).")
    elif rec_info is None:
        raise RuntimeError('Either the "input_dir" or the "rec_info" input argument should be set.')
    else:
        source_dir  = pathlib.Path(rec_info.source_directory)

    if rec_info is not None and not rec_info.source_directory:
        rec_info.source_directory = source_dir
    return source_dir, rec_info

def check_output_dir(output_dir: str|pathlib.Path, rec_info: Recording) -> tuple[pathlib.Path, Recording]:
    if output_dir is not None:
        output_dir  = pathlib.Path(output_dir)
        if rec_info is not None and rec_info.working_directory and pathlib.Path(rec_info.working_directory) != output_dir:
            raise ValueError(f"The provided output_dir ({output_dir}) does not equal the working directory set in rec_info ({rec_info.working_directory}).")
    elif rec_info is None:
        raise RuntimeError('Either the "output_dir" or the "rec_info" input argument should be set.')
    else:
        output_dir  = pathlib.Path(rec_info.working_directory)

    if output_dir.is_dir():
        with os.scandir(output_dir) as it:
            if any(it):
                raise RuntimeError(f'Output directory ({output_dir}) already exists and is not empty. Cannot use.')

    if rec_info is not None and not rec_info.working_directory:
        rec_info.working_directory = output_dir
    return output_dir, rec_info

def check_folders(output_dir: str|pathlib.Path, source_dir: str|pathlib.Path, rec_info: Recording, expected_eye_tracker: EyeTracker) -> tuple[pathlib.Path, pathlib.Path, Recording]:
    if rec_info is not None and rec_info.eye_tracker:
        if rec_info.eye_tracker!=expected_eye_tracker:
            raise ValueError(f'Provided rec_info is for a device ({rec_info.eye_tracker.value}) that is not a {expected_eye_tracker.value}. Cannot use.')

    source_dir, rec_info = check_source_dir(source_dir, rec_info)
    output_dir, rec_info = check_output_dir(output_dir, rec_info)
    return output_dir, source_dir, rec_info

def check_device(device: str|EyeTracker, rec_info: Recording):
    if device is None and (rec_info is None or not rec_info.eye_tracker):
        raise RuntimeError('Either the "device" or the eye_tracker field of the "rec_info" input argument should be set.')
    if device is not None:
        device = eyetracker.string_to_enum(device)
    if rec_info is not None and rec_info.eye_tracker:
        if device is not None:
            if rec_info.eye_tracker != device:
                raise ValueError(f'Provided device ({device.value}) did not match device specific in rec_info ({rec_info.eye_tracker.value}). Provide matching values or do not provide the device input argument.')
        else:
            device = eyetracker.string_to_enum(rec_info.eye_tracker)
    return device, rec_info


__all__ = ['pupil_core','pupil_invisible','pupil_neon','SeeTrue_STONE','SMI_ETG','tobii_G2','tobii_G3','adhawk_mindlink',
           'get_recording_info']


import numpy as np
import cv2
import pathlib

from . import data_files, drawing, ocv


class Gaze:
    # description of tsv file used for storage
    _columns_compressed = {'timestamp': 1, 'frame_idx':1,
                           'gaze_pos_vid': 2, 'gaze_pos_3d': 3, 'gaze_dir_l': 3, 'gaze_ori_l': 3, 'gaze_dir_r': 3, 'gaze_ori_r': 3}
    _non_float          = {'frame_idx': int}

    def __init__(self,
                 timestamp      : float,
                 frame_idx      : int,
                 gaze_pos_vid   : np.ndarray,
                 gaze_pos_3d    : np.ndarray = None,
                 gaze_dir_l     : np.ndarray = None,
                 gaze_ori_l     : np.ndarray = None,
                 gaze_dir_r     : np.ndarray = None,
                 gaze_ori_r     : np.ndarray = None):
        self.timestamp   : float        = timestamp
        self.frame_idx   : int          = frame_idx

        self.gaze_pos_vid: np.ndarray   = gaze_pos_vid      # gaze point on the scene video
        self.gaze_pos_3d : np.ndarray   = gaze_pos_3d       # gaze point in the world (often binocular gaze point)
        self.gaze_dir_l  : np.ndarray   = gaze_dir_l
        self.gaze_ori_l  : np.ndarray   = gaze_ori_l
        self.gaze_dir_r  : np.ndarray   = gaze_dir_r
        self.gaze_ori_r  : np.ndarray   = gaze_ori_r

    @staticmethod
    def readFromFile(fileName:str|pathlib.Path) -> tuple[dict[int,list['Gaze']], int]:
        return data_files.read_file(fileName,
                                    Gaze, False, False, True)

    def draw(self, img:np.ndarray, cameraParams:ocv.CameraParams=None, subPixelFac=1):
        drawing.openCVCircle(img, self.gaze_pos_vid, 8, (0,255,0), 2, subPixelFac)
        # draw 3D gaze point as well, usually coincides with 2D gaze point, but not always. E.g. the Adhawk MindLink may
        # apply a correction for parallax error to the projected gaze point using the vergence signal.
        if self.gaze_pos_3d is not None and cameraParams.has_intrinsics():
            camRot = np.zeros((1,3)) if cameraParams.rotation_vec is None else cameraParams.rotation_vec
            camPos = np.zeros((1,3)) if cameraParams.position     is None else cameraParams.position
            a = cv2.projectPoints(np.array(self.gaze_pos_3d).reshape(1,3),camRot,camPos,cameraParams.camera_mtx,cameraParams.distort_coeffs)[0][0][0]
            drawing.openCVCircle(img, a, 5, (0,255,255), -1, subPixelFac)


import numpy as np
import cv2
from typing import Any

from . import drawing, intervals, ocv, plane, timestamps


class ArUcoDetector():
    def __init__(self, ArUco_dict: cv2.aruco.Dictionary, params: dict[str, Any]):
        self._det_params                          = cv2.aruco.DetectorParameters()
        self._det_params.cornerRefinementMethod   = cv2.aruco.CORNER_REFINE_SUBPIX    # good default, user can override
        for p in params:
            setattr(self._det_params, p, params[p])

        self._dict                      = ArUco_dict
        self._det                       = cv2.aruco.ArucoDetector(self._dict, self._det_params)
        self._board: cv2.aruco.Board    = None

        self._camera_params             = ocv.CameraParams(None, None)

    def create_board(self, board_corner_points, ids):
        self._board = create_board(board_corner_points, ids, self._dict)

    def set_board(self, board):
        self._board = board

    def set_intrinsics(self, camera_params: ocv.CameraParams):
        self._camera_params = camera_params

    def detect_markers(self, image: cv2.UMat, min_nmarker_refine = 3):
        corners, ids, rejected_corners = self._det.detectMarkers(image)
        recovered_ids = None
        if np.any(ids==None):
            ids = None

        # Refine detected markers (eliminates markers not part of our poster, adds missing markers to the poster)
        if self._board and ids is not None and min_nmarker_refine and len(ids)>min_nmarker_refine:
            corners, ids, rejected_corners, recovered_ids = \
                self.refine_detected_markers(image, corners, ids, rejected_corners)
        return corners, ids, rejected_corners, recovered_ids

    def refine_detected_markers(self, image, detected_corners, detected_ids, rejected_corners):
        corners, ids, rejectedImgPoints, recoveredIds = self._det.refineDetectedMarkers(
                                image = image, board = self._board,
                                detectedCorners = detected_corners, detectedIds = detected_ids, rejectedCorners = rejected_corners,
                                cameraMatrix = self._camera_params.camera_mtx, distCoeffs = self._camera_params.distort_coeffs)
        if corners and corners[0].shape[0]==4:
            # there are versions out there where there is a bug in output shape of each set of corners, fix up
            corners = [np.reshape(c,(1,4,2)) for c in corners]
        if rejectedImgPoints and rejectedImgPoints[0].shape[0]==4:
            # same as for corners
            rejectedImgPoints = [np.reshape(c,(1,4,2)) for c in rejectedImgPoints]

        return corners, ids, rejectedImgPoints, recoveredIds

    def _match_image_points(self, corners, ids):
        return self._board.matchImagePoints(corners, ids) # -> objP, imgP

    def estimate_pose(self, corners, ids) -> tuple[int, np.ndarray, np.ndarray]:
        objP, imgP = self._match_image_points(corners, ids)
        return self._estimate_pose_impl(objP, imgP)

    def _estimate_pose_impl(self, objP, imgP):
        N_markers, R_vec, T_vec = 0, None, None
        if objP is None or not self._camera_params.has_intrinsics():
            return N_markers, R_vec, T_vec

        ok, R_vec, T_vec = cv2.solvePnP(objP, imgP, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, np.empty(1), np.empty(1))
        if ok:
            N_markers = int(objP.shape[0]/4)
        return N_markers, R_vec, T_vec

    def estimate_homography(self, corners, ids) -> tuple[np.ndarray, bool]:
        objP, imgP = self._match_image_points(corners, ids)
        return self._estimate_homography_impl(objP, imgP)

    def _estimate_homography_impl(self, objP, imgP):
        from . import transforms
        N_markers, H = 0, None
        if objP is None:
            return N_markers, H

        # use undistorted marker corners if possible
        if self._camera_params.has_intrinsics():
            imgP = np.vstack([cv2.undistortPoints(x, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, P=self._camera_params.camera_mtx) for x in imgP])

        H, status = transforms.estimateHomography(objP, imgP)
        if status:
            N_markers = int(objP.shape[0]/4)
        return N_markers, H

    # higher level functions for detecting + pose estimation, and for results visualization
    def detect_and_estimate(self, frame, frame_idx, min_num_markers) -> tuple[plane.Pose, dict[str,Any]]:
        pose = plane.Pose(frame_idx)
        corners, ids, rejectedImgPoints, recoveredIds = self.detect_markers(frame, min_nmarker_refine=min_num_markers)

        if ids is not None and len(ids) >= min_num_markers:
            # get matching image and board points
            objP, imgP = self._match_image_points(corners, ids)

            # get camera pose
            pose.pose_N_markers, pose.pose_R_vec, pose.pose_T_vec = \
                self._estimate_pose_impl(objP, imgP)

            # also get homography (direct image plane to plane in world transform)
            pose.homography_N_markers, pose.homography_mat = \
                self._estimate_homography_impl(objP, imgP)

        return pose, {'corners': corners, 'ids': ids, 'rejectedImgPoints': rejectedImgPoints, 'recoveredIds': recoveredIds}

    def visualize(self, frame, pose: plane.Pose, detect_dict, arm_length, sub_pixel_fac = 8, show_rejected_markers = False):
        if pose.pose_N_markers>0:
            # draw axis indicating poster pose (origin and orientation)
            drawing.openCVFrameAxis(frame, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, pose.pose_R_vec, pose.pose_T_vec, arm_length, 3, sub_pixel_fac)

        if pose.homography_N_markers>0:
            # find where plane origin is expected to be in the image
            target = pose.planeToCamHomography([0., 0.], self._camera_params)
            # draw target location on image
            if target[0] >= 0 and target[0] < frame.shape[1] and target[1] >= 0 and target[1] < frame.shape[0]:
                drawing.openCVCircle(frame, target, 3, (0,0,0), -1, sub_pixel_fac)

        # if any markers were detected, draw where on the frame
        drawing.arucoDetectedMarkers(frame, detect_dict['corners'], detect_dict['ids'], subPixelFac=sub_pixel_fac, specialHighlight=[detect_dict['recoveredIds'],(255,255,0)])

        # for debug, can draw rejected markers on frame
        if show_rejected_markers:
            cv2.aruco.drawDetectedMarkers(frame, detect_dict['rejectedImgPoints'], None, borderColor=(211,0,148))



def create_board(board_corner_points: list[np.ndarray], ids: list[int], ArUco_dict: cv2.aruco.Dictionary):
    board_corner_points = np.dstack(board_corner_points)        # list of 2D arrays -> 3D array
    board_corner_points = np.rollaxis(board_corner_points,-1)   # 4x2xN -> Nx4x2
    board_corner_points = np.pad(board_corner_points,((0,0),(0,0),(0,1)),'constant', constant_values=(0.,0.)) # Nx4x2 -> Nx4x3
    return cv2.aruco.Board(board_corner_points, ArUco_dict, np.array(ids))

def run_pose_estimation(in_video, frame_timestamp_file, camera_calibration_file,
                        output_dir, out_file,
                        processing_intervals,
                        ArUco_board, aruco_params, min_num_markers,
                        gui, arm_length, sub_pixel_fac = 8, show_rejected_markers = False):
    show_visualization = gui is not None

    # open video
    cap = ocv.CV2VideoReader(in_video, timestamps.from_file(frame_timestamp_file))

    # setup aruco marker detection
    detector = ArUcoDetector(ArUco_board.getDictionary(), aruco_params)
    detector.set_board(ArUco_board)
    # get camera calibration info
    detector.set_intrinsics(ocv.CameraParams.readFromFile(camera_calibration_file))


    stop_all_processing = False
    poses = []
    while True:
        # process frame-by-frame
        done, frame, frame_idx, frame_ts = cap.read_frame(report_gap=True)
        if done or intervals.beyond_last_interval(frame_idx, processing_intervals):
            break
        cap.report_frame()

        if frame is None:
            # we don't have a valid frame, continue to next
            continue

        if show_visualization:
            keys = gui.get_key_presses()
            if 'q' in keys:
                # quit fully
                stop_all_processing = True
                break
            if 'n' in keys:
                # goto next
                break

        # check we're in a current interval, else skip processing
        # NB: have to spool through like this, setting specific frame to read
        # with cap.set(cv2.CAP_PROP_POS_FRAMES) doesn't seem to work reliably
        # for VFR video files
        if not intervals.is_in_interval(frame_idx, processing_intervals):
            continue

        # detect markers
        pose, detect_dict = detector.detect_and_estimate(frame, frame_idx, min_num_markers=min_num_markers)
        poses.append(pose)
        # draw detection and pose, if wanted
        if show_visualization:
            detector.visualize(frame, pose, detect_dict, arm_length, sub_pixel_fac, show_rejected_markers)

        if show_visualization:
            # keys is populated above
            if 's' in keys:
                # screenshot
                cv2.imwrite(str(output_dir / ('detect_frame_%d.png' % frame_idx)), frame)
            gui.update_image(frame, frame_ts/1000., frame_idx)
            closed, = gui.get_state()
            if closed:
                stop_all_processing = True
                break

    plane.Pose.writeToFile(poses, output_dir / out_file, skip_failed=True)

    if show_visualization:
        gui.stop()

    return stop_all_processing


import numpy as np

class Marker:
    def __init__(self, key, center, corners=None, color=None, rot=0):
        self.key = key
        self.center = center
        self.corners = corners
        self.color = color
        self.rot = rot

    def __str__(self):
        ret = '[%d]: center @ (%.2f, %.2f), rot %.0f deg' % (self.key, self.center[0], self.center[1], self.rot)
        return ret

def corners_intersection(corners):
    line1 = ( corners[0], corners[2] )
    line2 = ( corners[1], corners[3] )
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return np.array( [x,y] ).astype('float32')

def getUnrotated(cornerPoints, rot):
    # markers are rotated in multiples of 90 only, so can easily unrotate
    if rot == -90:
        # -90 deg
        cornerPoints = np.vstack((cornerPoints[-1,:], cornerPoints[0:3,:]))
    elif rot == 90:
        # 90 deg
        cornerPoints = np.vstack((cornerPoints[1:,:], cornerPoints[0,:]))
    elif rot == 180:
        # 180 deg
        cornerPoints = np.vstack((cornerPoints[2:,:], cornerPoints[0:2,:]))

    return cornerPoints


def is_in_interval(frame_idx, intervals):
    if intervals is None:
        return True # no interval defined, that means all frames

    # return True if we're in a current interval
    for f in range(0,len(intervals),2):
        if frame_idx>=intervals[f] and frame_idx<=intervals[f+1]:
            return True
    return False

def beyond_last_interval(frame_idx, intervals):
    if intervals is not None and frame_idx > intervals[-1]:
        return True
    return False


import numpy as np
import pandas as pd
import cv2
import pathlib
import math

from . import data_files, drawing, marker, ocv


class Plane:
    default_ref_image_name = 'reference_image.png'

    def __init__(self,
                 markers                : str|pathlib.Path|pd.DataFrame,                            # if str or Path: file from which to read markers. Else direction N_markerx4 array. Should contain centers of markers
                 marker_size            : float,                                                    # in "unit" units

                 aruco_dict                                             = cv2.aruco.DICT_4X4_250,
                 marker_border_bits                                     = 1,
                 marker_pos_scale_fac                                   = 1.,                       # scale factor for marker positions in the markers input argument
                 unit                   : str                           = None,                     # Unit in which measurements (marker size and positions for instance) are expressed. Purely informational
                 package_to_read_from   : str                           = None,                     # if provided, reads marker file from specified package's resources
                 plane_center           : np.ndarray                    = None,                     # center coordinate of plane, in scaled marker coordinates
                 ref_image_store_path   : str|pathlib.Path              = None,
                 ref_image_width                                        = 1920
                 ):

        self.marker_size                                    = marker_size
        # marker positions
        self.markers            : dict[int,marker.Marker]   = {}
        self.bbox               : list[float]               = []

        # marker specs
        self.aruco_dict                                     = cv2.aruco.getPredefinedDictionary(aruco_dict)
        self.marker_border_bits                             = marker_border_bits
        self.unit                                           = unit

        # prep markers
        self._get_markers(markers, marker_pos_scale_fac, package_to_read_from, plane_center)

        # get reference image of plane
        if ref_image_store_path:
            ref_image_store_path = pathlib.Path(ref_image_store_path)

        # get image
        img = None
        # read from file if image exists
        if ref_image_store_path is not None and ref_image_store_path.is_file():
            img = cv2.imread(str(ref_image_store_path), cv2.IMREAD_COLOR)
        # if image doesn't exist or is the wrong size, create
        if img is None or img.shape[1]!=ref_image_width:
            img = self._store_reference_image(ref_image_store_path, ref_image_width)

        self._ref_image_width                               = ref_image_width
        self._ref_image_cache   : dict[int, np.ndarray]     = {ref_image_width: img}

    def get_ref_image(self, width: int=None, asRGB=False):
        if width is None:
            width = self._ref_image_width
        assert isinstance(width,int), f'width input should be an int, not {type(width)}'
        # check we have the image, if not, add to cache
        if width not in self._ref_image_cache:
            scale = float(width)/self._ref_image_width
            self._ref_image_cache[width] = cv2.resize(self._ref_image_cache[self._ref_image_width], None, fx=scale, fy=scale, interpolation = cv2.INTER_AREA)

        # return
        if asRGB:
            return self._ref_image_cache[width][:,:,[2,1,0]]    # indexing returns a copy
        else:
            # OpenCV's BGR
            return self._ref_image_cache[width].copy()

    def draw(self, img, x, y, subPixelFac=1, color=None, size=6):
        from . import transforms
        if not math.isnan(x):
            xy = transforms.toImagePos(x, y, self.bbox, [img.shape[1], img.shape[0]])
            if color is None:
                drawing.openCVCircle(img, xy, 8, (0,255,0), -1, subPixelFac)
                color = (0,0,0)
            drawing.openCVCircle(img, xy, size, color, -1, subPixelFac)

    def _get_markers(self, markers: str|pathlib.Path|pd.DataFrame, marker_pos_scale_fac: float, package_to_read_from: str|None, plane_center: np.ndarray|None):
        # read in aruco marker positions
        markerHalfSizeMm  = self.marker_size/2.

        if isinstance(markers, pd.DataFrame):
            marker_pos = markers
        else:
            marker_pos = data_files.read_coord_file(markers, package_to_read_from)
        if marker_pos is None:
            return

        # turn into marker objects
        marker_pos.x = marker_pos.x.astype('float32')*marker_pos_scale_fac
        marker_pos.y = marker_pos.y.astype('float32')*marker_pos_scale_fac
        if plane_center is not None:
            marker_pos.x -= plane_center[0]
            marker_pos.y -= plane_center[1]

        for idx, row in marker_pos.iterrows():
            c   = row[['x','y']].values
            # rotate markers (negative because plane coordinate system)
            rot = row[['rotation_angle']].values[0]
            if rot%90 != 0:
                raise ValueError("Rotation of a marker must be a multiple of 90 degrees")
            rotr= -math.radians(rot)
            R   = np.array([[math.cos(rotr), math.sin(rotr)], [-math.sin(rotr), math.cos(rotr)]])
            # top left first, and clockwise: same order as detected ArUco marker corners
            tl = c + np.matmul(R, np.array([-markerHalfSizeMm, -markerHalfSizeMm]))
            tr = c + np.matmul(R, np.array([ markerHalfSizeMm, -markerHalfSizeMm]))
            br = c + np.matmul(R, np.array([ markerHalfSizeMm,  markerHalfSizeMm]))
            bl = c + np.matmul(R, np.array([-markerHalfSizeMm,  markerHalfSizeMm]))

            self.markers[idx] = marker.Marker(idx, c, corners=[tl, tr, br, bl], rot=rot)

        # determine bounding box of markers ([left, top, right, bottom])
        # NB: this assumes markers are rotated by multiples of 90 degrees
        self.bbox.append(marker_pos.x.min()-markerHalfSizeMm)
        self.bbox.append(marker_pos.y.min()-markerHalfSizeMm)
        self.bbox.append(marker_pos.x.max()+markerHalfSizeMm)
        self.bbox.append(marker_pos.y.max()+markerHalfSizeMm)

    def get_aruco_board(self, unrotate_markers=False):
        from . import aruco
        board_corner_points = []
        ids = []
        for key in self.markers:
            ids.append(key)
            marker_corner_points = np.vstack(self.markers[key].corners).astype('float32')
            if unrotate_markers:
                marker_corner_points = marker.getUnrotated(marker_corner_points, self.markers[key].rot)

            board_corner_points.append(marker_corner_points)
        return aruco.create_board(board_corner_points, ids, self.aruco_dict)

    def _store_reference_image(self, path: pathlib.Path, width: int) -> np.ndarray:
        referenceBoard = self.get_aruco_board(unrotate_markers = True)
        # get image with markers
        bboxExtents = [self.bbox[2]-self.bbox[0], math.fabs(self.bbox[3]-self.bbox[1])]  # math.fabs to deal with bboxes where (-,-) is bottom left
        aspectRatio = bboxExtents[0]/bboxExtents[1]
        height      = math.ceil(width/aspectRatio)
        margin      = 1  # always 1 pixel, anything else behaves strangely (markers are drawn over margin as well)

        img  = cv2.cvtColor(
            referenceBoard.generateImage(
                (width+2*margin,height+2*margin),margin,self.marker_border_bits),
            cv2.COLOR_GRAY2RGB
        )
        # cut off this 1-pix margin
        assert img.shape[0]==height+2*margin,"Output image height is not as expected"
        assert img.shape[1]==width +2*margin,"Output image width is not as expected"
        img  = img[1:-1,1:-1,:]
        # walk through all markers, if any are supposed to be rotated, do so
        minX =  np.inf
        maxX = -np.inf
        minY =  np.inf
        maxY = -np.inf
        rots = []
        corner_points_unrot = []
        for key in self.markers:
            corner_points = np.vstack(self.markers[key].corners).astype('float32')
            corner_points_unrot.append(marker.getUnrotated(corner_points, self.markers[key].rot))
            rots.append(self.markers[key].rot)
            minX = np.min(np.hstack((minX,corner_points[:,0])))
            maxX = np.max(np.hstack((maxX,corner_points[:,0])))
            minY = np.min(np.hstack((minY,corner_points[:,1])))
            maxY = np.max(np.hstack((maxY,corner_points[:,1])))
        if np.any(np.array(rots)!=0):
            # determine where the markers are placed
            sizeX = maxX - minX
            sizeY = maxY - minY
            xReduction = sizeX / float(img.shape[1])
            yReduction = sizeY / float(img.shape[0])
            if xReduction > yReduction:
                nRows = int(sizeY / xReduction);
                yMargin = (img.shape[0] - nRows) / 2;
                xMargin = 0
            else:
                nCols = int(sizeX / yReduction);
                xMargin = (img.shape[1] - nCols) / 2;
                yMargin = 0

            for r,cpu in zip(rots,corner_points_unrot):
                if r != 0:
                    # figure out where marker is
                    cpu -= np.array([[minX,minY]])
                    cpu[:,0] =       cpu[:,0] / sizeX  * float(img.shape[1]) + xMargin
                    cpu[:,1] = (1. - cpu[:,1] / sizeY) * float(img.shape[0]) + yMargin
                    sz = np.min(cpu[2,:]-cpu[0,:])
                    # get marker
                    cpu = np.floor(cpu)
                    idxs = np.floor([cpu[0,1], cpu[0,1]+sz, cpu[0,0], cpu[0,0]+sz]).astype('int')
                    mark = img[idxs[0]:idxs[1], idxs[2]:idxs[3]]
                    # rotate (opposite because coordinate system) and put back
                    if r==-90:
                        mark = cv2.rotate(mark, cv2.ROTATE_90_CLOCKWISE)
                    elif r==90:
                        mark = cv2.rotate(mark, cv2.ROTATE_90_COUNTERCLOCKWISE)
                    elif r==180:
                        mark = cv2.rotate(mark, cv2.ROTATE_180)

                    img[idxs[0]:idxs[1], idxs[2]:idxs[3]] = mark

        if path:
            cv2.imwrite(str(path), img)

        return img


class Pose:
    # description of tsv file used for storage
    _columns_compressed = {'frame_idx':1,
                           'pose_N_markers': 1, 'pose_R_vec': 3, 'pose_T_vec': 3,
                           'homography_N_markers': 1, 'homography_mat': 9}
    _non_float          = {'frame_idx': int, 'pose_ok': bool, 'pose_N_markers': int, 'homography_N_markers': int}

    def __init__(self,
                 frame_idx              : int,
                 pose_N_markers         : int       = 0,
                 pose_R_vec             : np.ndarray= None,
                 pose_T_vec             : np.ndarray= None,
                 homography_N_markers   : int       = 0,
                 homography_mat         : np.ndarray= None):
        self.frame_idx            : int         = frame_idx
        # pose
        self.pose_N_markers       : int         = pose_N_markers        # number of ArUco markers this pose estimate is based on. 0 if failed
        self.pose_R_vec           : np.ndarray  = pose_R_vec
        self.pose_T_vec           : np.ndarray  = pose_T_vec
        # homography
        self.homography_N_markers : int         = homography_N_markers  # number of ArUco markers this homongraphy estimate is based on. 0 if failed
        self.homography_mat       : np.ndarray  = homography_mat.reshape(3,3) if homography_mat is not None else homography_mat

        # internals
        self._RMat        = None
        self._RtMat       = None
        self._planeNormal = None
        self._planePoint  = None
        self._RMatInv     = None
        self._RtMatInv    = None
        self._iH          = None

    @staticmethod
    def readFromFile(fileName:str|pathlib.Path, start:int=None, end:int=None) -> dict[int,'Pose']:
        return data_files.read_file(fileName,
                                    Pose, True, True, False,
                                    start=start, end=end)[0]

    @staticmethod
    def writeToFile(poses: list['Pose'], fileName:str|pathlib.Path, skip_failed=False):
        data_files.write_array_to_file(poses, fileName,
                                       Pose._columns_compressed,
                                       skip_all_nan=skip_failed)

    def camToWorld(self, point: np.ndarray):
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)):
            return np.full((3,), np.nan)

        if self._RtMatInv is None:
            if self._RMatInv is None:
                if self._RMat is None:
                    self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
                self._RMatInv = self._RMat.T
            self._RtMatInv = np.hstack((self._RMatInv,np.matmul(-self._RMatInv,self.pose_T_vec.reshape(3,1))))

        return np.matmul(self._RtMatInv,np.append(np.array(point),1.).reshape((4,1))).flatten()

    def worldToCam(self, point: np.ndarray):
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)):
            return np.full((3,), np.nan)

        if self._RtMat is None:
            if self._RMat is None:
                self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
            self._RtMat = np.hstack((self._RMat, self.pose_T_vec.reshape(3,1)))

        return np.matmul(self._RtMat,np.append(np.array(point),1.).reshape((4,1))).flatten()

    def planeToCamHomography(self, point: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on plane (2D) to location on camera image (2D)
        from . import transforms
        if self.homography_mat is None:
            return np.full((2,), np.nan)

        if self._iH is None:
            self._iH = np.linalg.inv(self.homography_mat)
        out = transforms.applyHomography(self._iH, *point)
        if camera_params.has_intrinsics():
            out = transforms.distortPoint(*out, camera_params.camera_mtx, camera_params.distort_coeffs)
        return out

    def camToPlaneHomography(self, point: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on camera image (2D) to location on plane (2D)
        from . import transforms
        if self.homography_mat is None:
            return np.full((2,), np.nan)

        if camera_params.has_intrinsics():
            point = transforms.undistortPoint(*point, camera_params.camera_mtx, camera_params.distort_coeffs)
        return transforms.applyHomography(self.homography_mat, *point)

    def getOriginOnImage(self, camera_params: ocv.CameraParams) -> np.ndarray:
        if self.pose_N_markers>0 and camera_params.has_intrinsics():
            a = cv2.projectPoints(np.zeros((1,3)), self.pose_R_vec,self.pose_T_vec, camera_params.camera_mtx,camera_params.distort_coeffs)[0].flatten()
        elif self.homography_N_markers>0:
            a = self.planeToCamHomography([0., 0.], camera_params.camera_mtx, camera_params.distort_coeffs)
        else:
            a = np.full((2,), np.nan)
        return a

    def vectorIntersect(self, vector: np.ndarray, origin = np.array([0.,0.,0.])):
        from . import transforms

        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(vector)):
            return np.full((3,), np.nan)

        if self._planeNormal is None:
            if self._RtMat is None:
                if self._RMat is None:
                    self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
                self._RtMat = np.hstack((self._RMat, self.pose_T_vec.reshape(3,1)))

            # get poster normal
            self._planeNormal = np.matmul(self._RMat, np.array([0., 0., 1.]))
            # get point on poster (just use origin)
            self._planePoint  = np.matmul(self._RtMat, np.array([0., 0., 0., 1.]))

        # normalize vector
        vector /= np.sqrt((vector**2).sum())

        # find intersection of 3D gaze with poster
        return transforms.intersect_plane_ray(self._planeNormal, self._planePoint, vector.flatten(), origin.flatten())


import enum
import json
import pathlib
import dataclasses
import typing
import numpy as np


def hex_to_rgba_0_1(hex):
    r = int(hex[1:3], base=16) / 255
    g = int(hex[3:5], base=16) / 255
    b = int(hex[5:7], base=16) / 255
    if len(hex) > 7:
        a = int(hex[7:9], base=16) / 255
    else:
        a = 1.0
    return (r, g, b, a)

def rgba_0_1_to_hex(rgba):
    r = "%.2x" % int(rgba[0] * 255)
    g = "%.2x" % int(rgba[1] * 255)
    b = "%.2x" % int(rgba[2] * 255)
    if len(rgba) > 3:
        a = "%.2x" % int(rgba[3] * 255)
    else:
        a = "FF"
    return f"#{r}{g}{b}{a}"


class AutoName(enum.Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name.strip("_").replace("__", "-").replace("_", " ")


@dataclasses.dataclass
class CustomTypeEntry:
    type        : typing.Type
    reg_name    : str
    to_json     : typing.Callable
    from_json   : typing.Callable

CUSTOM_TYPE_REGISTRY = []
def register_type(entry: CustomTypeEntry):
    CUSTOM_TYPE_REGISTRY.append(entry)
register_type(CustomTypeEntry(pathlib.Path,'pathlib.Path',str,lambda x: pathlib.Path(x)))

class CustomTypeEncoder(json.JSONEncoder):
    def default(self, obj):
        for t in CUSTOM_TYPE_REGISTRY:
            if isinstance(obj, t.type):
                return {t.reg_name: t.to_json(obj)}
        return json.JSONEncoder.default(self, obj)

def json_reconstitute(d):
    for t in CUSTOM_TYPE_REGISTRY:
        if t.reg_name in d:
            return t.from_json(d[t.reg_name])
    return d


def cartesian_product(*arrays):
    ndim = len(arrays)
    return (np.stack(np.meshgrid(*arrays), axis=-1).reshape(-1, ndim))



import numpy as np
import math
import cv2

from . import marker

def openCVCircle(img, center_coordinates, radius, color, thickness, subPixelFac):
    p = [np.round(x*subPixelFac) for x in center_coordinates]
    if np.all([not math.isnan(x) and abs(x)<np.iinfo(int).max for x in p]):
        p = tuple([int(x) for x in p])
        cv2.circle(img, p, radius*subPixelFac, color, thickness, lineType=cv2.LINE_AA, shift=int(math.log2(subPixelFac)))

def openCVLine(img, start_point, end_point, color, thickness, subPixelFac):
    sp = [np.round(x*subPixelFac) for x in start_point]
    ep = [np.round(x*subPixelFac) for x in   end_point]
    if np.all([not math.isnan(x) and abs(x)<np.iinfo(int).max for x in sp]) and np.all([not math.isnan(x) and abs(x)<np.iinfo(int).max for x in ep]):
        sp = tuple([int(x) for x in sp])
        ep = tuple([int(x) for x in ep])
        cv2.line(img, sp, ep, color, thickness, lineType=cv2.LINE_AA, shift=int(math.log2(subPixelFac)))

def openCVRectangle(img, p1, p2, color, thickness, subPixelFac):
    p1 = [np.round(x*subPixelFac) for x in p1]
    p2 = [np.round(x*subPixelFac) for x in p2]
    if np.all([not math.isnan(x) and abs(x)<np.iinfo(int).max for x in p1]) and np.all([not math.isnan(x) and abs(x)<np.iinfo(int).max for x in p2]):
        p1 = tuple([int(x) for x in p1])
        p2 = tuple([int(x) for x in p2])
        cv2.rectangle(img, p1, p2, color, thickness, lineType=cv2.LINE_AA, shift=int(math.log2(subPixelFac)))

def openCVFrameAxis(img, cameraMatrix, distCoeffs, rvec,  tvec,  armLength, thickness, subPixelFac, position = [0.,0.,0.]):
    # same as the openCV function, but with anti-aliasing for a nicer image if subPixelFac>1
    points = np.vstack((np.zeros((1,3)), armLength*np.eye(3)))+np.vstack(4*[np.asarray(position)])
    points = cv2.projectPoints(points, rvec, tvec, cameraMatrix, distCoeffs)[0]
    openCVLine(img, points[0].flatten(), points[1].flatten(), (0, 0, 255), thickness, subPixelFac)
    openCVLine(img, points[0].flatten(), points[2].flatten(), (0, 255, 0), thickness, subPixelFac)
    openCVLine(img, points[0].flatten(), points[3].flatten(), (255, 0, 0), thickness, subPixelFac)

def arucoDetectedMarkers(img,corners,ids,borderColor=(0,255,0), drawIDs = True, subPixelFac=1, specialHighlight = []):
    # same as the openCV function, but with anti-aliasing for a (much) nicer image if subPixelFac>1
    textColor   = [x for x in borderColor]
    cornerColor = [x for x in borderColor]
    textColor[0]  , textColor[1]   = textColor[1]  , textColor[0]       #   text color just swap B and R
    cornerColor[1], cornerColor[2] = cornerColor[2], cornerColor[1]     # corner color just swap G and R

    drawIDs = drawIDs and (ids is not None) and len(ids)>0

    for i in range(0, len(corners)):
        corner = corners[i][0]
        # draw marker sides
        sideColor = borderColor
        for s,c in zip(specialHighlight[::2],specialHighlight[1::2]):
            if s is not None and ids[i][0] in s:
                sideColor = c
        for j in range(4):
            p0 = corner[j,:]
            p1 = corner[(j + 1) % 4,:]
            openCVLine(img, p0, p1, sideColor, 1, subPixelFac)

        # draw first corner mark
        p1 = corner[0]
        openCVRectangle(img, corner[0]-3, corner[0]+3, cornerColor, 1, subPixelFac)

        # draw IDs if wanted
        if drawIDs:
            c = marker.corners_intersection(corner)
            cv2.putText(img, str(ids[i][0]), tuple(c.astype('int')), cv2.FONT_HERSHEY_SIMPLEX, 0.6, textColor, 2, lineType=cv2.LINE_AA)


import numpy as np
import pandas as pd
import cv2

from .mp4analyser import iso


def av_rescale(a, b, c):
    # a * b / c, rounding to nearest and halfway cases away from zero
    # e.g., scale a from timebase c to timebase b
    # from ffmpeg libavutil mathematics.c, porting the simple case assuming that a, b and c <= INT_MAX
    r = c // 2
    return (a * b + r) // c


def getFrameTimestampsFromVideo(vid_file):
    """
    Parse the supplied video, return an array of frame timestamps. There must be only one video stream
    in the video file, because otherwise we do not know which is the correct stream.
    """
    if vid_file.suffix in ['.mov', '.mp4', '.m4a', '.3gp', '.3g2', '.mj2']:
        # parse mp4 file
        boxes       = iso.Mp4File(str(vid_file))
        summary     = boxes.get_summary()
        vid_tracks  = [t for t in summary['track_list'] if t['media_type']=='video']
        assert len(vid_tracks)==1, f"File has {len(vid_tracks)} video tracks (more than one), not supported"
        # 1. find mdat box
        moov        = boxes.children[[i for i,x in enumerate(boxes.children) if x.type=='moov'][0]]
        # 2. get global/movie time scale
        movie_time_scale = np.int64(moov.children[[i for i,x in enumerate(moov.children) if x.type=='mvhd'][0]].box_info['timescale'])
        # 3. find video track boxes
        trak_idxs   = [i for i,x in enumerate(moov.children) if x.type=='trak']
        trak_idxs   = [x for i,x in enumerate(trak_idxs) if summary['track_list'][i]['media_type']=='video']
        assert len(trak_idxs)==1
        trak        = moov.children[trak_idxs[0]]
        # 4. get mdia box
        mdia        = trak.children[[i for i,x in enumerate(trak.children) if x.type=='mdia'][0]]
        # 5. get media/track time_scale and duration fields from mdhd
        mdhd            = mdia.children[[i for i,x in enumerate(mdia.children) if x.type=='mdhd'][0]]
        media_time_scale= mdhd.box_info['timescale']
        # 6. check for presence of edit list
        # if its there, check its one we understand (one or multiple empty list at the beginning
        # to shift movie start, and/or a single non-empty edit list), and parse it. Else abort
        edts_idx= [i for i,x in enumerate(trak.children) if x.type=='edts']
        empty_duration  = np.int64(0)
        media_start_time= np.int64(-1)
        media_duration  = np.int64(-1)
        if edts_idx:
            elst = trak.children[edts_idx[0]].children[0]
            edit_start_index = 0
            # logic ported from mov_build_index()/mov_fix_index() in ffmpeg's libavformat/mov.c
            for i in range(elst.box_info['entry_count']):
                if i==edit_start_index and elst.box_info['entry_list'][i]['media_time'] == -1:
                    # empty edit list, indicates the start time of the stream
                    # relative to the presentation itself
                    this_empty_duration  = np.int64(elst.box_info['entry_list'][i]['segment_duration'])  # NB: in movie time scale
                    # convert duration from edit list from global timescale to track timescale
                    empty_duration  += av_rescale(this_empty_duration,media_time_scale,movie_time_scale)
                    edit_start_index+= 1
                elif i==edit_start_index and elst.box_info['entry_list'][i]['media_time'] > 0:
                    media_start_time = np.int64(elst.box_info['entry_list'][i]['media_time'])   # NB: already in track timescale, do not scale
                    media_duration   = av_rescale(np.int64(elst.box_info['entry_list'][i]['segment_duration']),media_time_scale,movie_time_scale)   # as above, scale to track timescale
                    if media_start_time<0:
                        raise RuntimeError('File contains an edit list that is too complicated (media start time < 0) for this parser, not supported')
                    if elst.box_info['entry_list'][i]['media_rate']!=1.0:
                        raise RuntimeError('File contains an edit list that is too complicated (media time is not 1.0) for this parser, not supported')
                elif i>edit_start_index:
                    raise RuntimeError('File contains an edit list that is too complicated (multiple non-empty edits) for this parser, not supported')
        # 7. get stbl
        minf    = mdia.children[[i for i,x in enumerate(mdia.children) if x.type=='minf'][0]]
        stbl    = minf.children[[i for i,x in enumerate(minf.children) if x.type=='stbl'][0]]
        # 8. check whether we have a ctts atom
        ctts_idx= [i for i,x in enumerate(stbl.children) if x.type=='ctts']
        if ctts_idx:
            ctts_ = stbl.children[ctts_idx[0]].box_info['entry_list']
            if any([e['sample_offset']<0 for e in ctts_]):
                # if we need to deal with them, we'd probably want to completely port mov_build_index() and mov_fix_index() in ffmpeg's libavformat/mov.c
                raise RuntimeError('Encountered a ctts (composition offset) atom with negative sample offsets, cannot handle that. aborting.')
            # uncompress
            total_frames_ctts = sum([e['sample_count'] for e in ctts_])
            ctts = np.zeros(total_frames_ctts, dtype=np.int64)
            idx = 0
            for e in ctts_:
                ctts[idx:idx+e['sample_count']] = e['sample_offset']
                idx = idx+e['sample_count']
        # 9. get sample table from stts
        stts = stbl.children[[i for i,x in enumerate(stbl.children) if x.type=='stts'][0]].box_info['entry_list']
        # uncompress the delta table
        total_frames_stts = sum([e['sample_count'] for e in stts])
        dts = np.zeros(total_frames_stts, dtype=np.int64)
        idx = 0
        for e in stts:
            dts[idx:idx+e['sample_count']] = e['sample_delta']
            idx = idx+e['sample_count']

        # 10. now put it all together
        # turn into timestamps
        dts = np.cumsum(np.insert(dts, 0, 0))
        dts = np.delete(dts,-1) # remove last, since that denotes _end_ of last frame, and we only need timestamps for frame onsets
        # apply ctts
        if ctts_idx:
            dts += ctts
            # ctts should lead to a reordering of frames, so sort
            dts = np.sort(dts)
        # if we have a non-empty edit list, apply
        if media_start_time!=-1:
            # remove all timestamps before start or after end of edit list
            to_keep = np.logical_or(dts >= media_start_time, dts < (media_duration + media_start_time))
            dts = dts[to_keep]
            min_corrected_pts = dts[0]  # already sorted, this is the first frame's pts
            # If there are empty edits, then min_corrected_pts might be positive
            # intentionally. So we subtract the sum duration of emtpy edits here.
            min_corrected_pts -= empty_duration
            # If the minimum pts turns out to be greater than zero,
            # then we subtract the dts by that amount to make the first pts zero.
            dts -= min_corrected_pts
        # now turn into timestamps in ms
        frameTs = (dts+empty_duration)/media_time_scale*1000
    else:
        # open file with opencv and get timestamps of each frame
        vid = cv2.VideoCapture(str(vid_file))
        nframes = float(vid.get(cv2.CAP_PROP_FRAME_COUNT))
        frameTs = []
        frame_idx = 0
        while vid.isOpened():
            # read timestamp _before_ reading the frame, so we get position at start of the frame, not at
            # end
            ts = vid.get(cv2.CAP_PROP_POS_MSEC)

            ret, _ = vid.read()
            frame_idx += 1

            # You'd think to get the time at the start of the frame, which is what we want, you'd need to
            # read the time _before_ reading the frame. But there seems to be an off-by-one here for some
            # files, like at least some MP4s, but not in some AVIs in my testing. Catch this off-by-one
            # and to correct for it, do not store the first timestamp. This ensures we get a sensible
            # output (else first two frames have timestamp 0.0 ms).
            if frame_idx==1 and ts==vid.get(cv2.CAP_PROP_POS_MSEC):
                continue

            frameTs.append(ts)
            # check if we're done. Can't trust ret==False to indicate we're at end of video, as it may also return false for some frames when video has errors in the middle that we can just read past
            if (not ret and frame_idx>0 and frame_idx/nframes<.99):
                raise RuntimeError("The video file is corrupt. Testing has shown that it cannot be guaranteed that timestamps remain correct when trying to read past the hole. So abort, cannot process this video.")

        # release the video capture object
        vid.release()
        frameTs = np.array(frameTs)

    ### convert the frame_timestamps to dataframe
    frameIdx = np.arange(0, len(frameTs))
    frameTsDf = pd.DataFrame({'frame_idx': frameIdx, 'timestamp': frameTs})
    frameTsDf.set_index('frame_idx', inplace=True)

    return frameTsDf


def tssToFrameNumber(ts,frameTimestamps,mode='nearest'):
    df = pd.DataFrame(index=ts)
    df.insert(0,'frame_idx',np.int64(0))

    # get index where this ts would be inserted into the frame_timestamp array
    idxs = np.searchsorted(frameTimestamps, ts)
    if mode=='after':
        idxs = idxs.astype('float32')
        # out of range, set to nan
        idxs[idxs==0] = np.nan
        # -1: since idx points to frame timestamp for frame after the one during which the ts ocurred, correct
        idxs -= 1
    elif mode=='nearest':
        # implementation from https://stackoverflow.com/questions/8914491/finding-the-nearest-value-and-return-the-index-of-array-in-python/8929827#8929827
        # same logic as used by pupil labs
        idxs = np.clip(idxs, 1, len(frameTimestamps)-1)
        left = frameTimestamps[idxs-1]
        right = frameTimestamps[idxs]
        idxs -= ts - left < right - ts

    df=df.assign(frame_idx=idxs)
    if mode=='after':
        df=df.convert_dtypes() # turn into int64 again

    return df